/*
 * display.c
 *
 * Copyright 2019 Thomas Gollmer <th_goso@freenet.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 * Displayanstuerung und Textaugaben
 */

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "rtc.h"
#include "mp3player.h"
#include "buttons.h"
#include "eemem.h"
#include "light.h"
#include "display.h"

// Display Signalleitungen Data7...Data4 -> PC3...PC0
#define DISP_SET_E_HIGH       (PORTB |= (1<<PB0))
#define DISP_SET_E_LOW        (PORTB &= ~(1<<PB0))
#define DISP_SET_RS_HIGH      (PORTB |= (1<<PB2))
#define DISP_SET_RS_LOW       (PORTB &= ~(1<<PB2))

#define DISP_SET_D4_HIGH      (PORTC |= (1<<PC0))
#define DISP_SET_D4_LOW       (PORTC &= ~(1<<PC0))
#define DISP_SET_D5_HIGH      (PORTC |= (1<<PC1))
#define DISP_SET_D5_LOW       (PORTC &= ~(1<<PC1))
#define DISP_SET_D6_HIGH      (PORTC |= (1<<PC2))
#define DISP_SET_D6_LOW       (PORTC &= ~(1<<PC2))
#define DISP_SET_D7_HIGH      (PORTC |= (1<<PC3))
#define DISP_SET_D7_LOW       (PORTC &= ~(1<<PC3))


// Display Befehle
#define LCD_CLEAR_DISPLAY     0x01
#define LCD_CURSOR_HOME       0x02
#define LCD_SET_ENTRY         0x04
#define LCD_ENTRY_DECREASE    0x00
#define LCD_ENTRY_INCREASE    0x02
#define LCD_ENTRY_NOSHIFT     0x00
#define LCD_ENTRY_SHIFT       0x01
#define LCD_SET_DISPLAY       0x08
#define LCD_DISPLAY_OFF       0x00
#define LCD_DISPLAY_ON        0x04
#define LCD_CURSOR_OFF        0x00
#define LCD_CURSOR_ON         0x02
#define LCD_BLINKING_OFF      0x00
#define LCD_BLINKING_ON       0x01
#define LCD_SET_SHIFT         0x10
#define LCD_CURSOR_MOVE       0x00
#define LCD_DISPLAY_SHIFT     0x08
#define LCD_SHIFT_LEFT        0x00
#define LCD_SHIFT_RIGHT       0x04
#define LCD_SET_FUNCTION      0x20
#define LCD_FUNCTION_4BIT     0x00
#define LCD_FUNCTION_8BIT     0x10
#define LCD_FUNCTION_1LINE    0x00
#define LCD_FUNCTION_2LINE    0x08
#define LCD_FUNCTION_5X7      0x00
#define LCD_FUNCTION_5X10     0x04
#define LCD_SOFT_RESET        0x30
#define LCD_SET_CGADR         0x40
#define LCD_SET_DDADR         0x80
#define LCD_DDADR_LINE1       0x00
#define LCD_DDADR_LINE2       0x40
#define LCD_DDADR_LINE3       0x10
#define LCD_DDADR_LINE4       0x50

// Kleine Umlaute (große gibt's nicht)
#define UML_a           "\xE1"
#define UML_o           "\xEF"
#define UML_u           "\xF5"
#define UML_s           "\xE2"
// ---------------------------------------------------------------------------------------------------------------------
// Aktuelle Displaydaten --> wann auch immer normal beschreiben
// Wird im IRQ NUR gelesen und auf Display angezeigt
volatile char Display_Data[32] = {"Thomas  Gollmers  Licht-Wecker  "};
// ---------------------------------------------------------------------------------------------------------------------
// Displaytextbausteine IMMER aufgefüllt auf 16 Zeichen (für eine Zeile)
static const __flash char txt_moment[]              = "Moment bitte !  ";
static const __flash char txt_starte_mp3modul[]     = "Starte MP3-Modul";
static const __flash char txt_tracks_in_root[]      = "ROOT Tracks     ";
static const __flash char txt_spiele_nr1_an[]       = "Spiele  Nr.1  an";
static const __flash char txt_uhrzeit[]             = "Uhrzeit         ";
static const __flash char txt_alarmzeit[]           = "Alarmzeit       ";
static const __flash char txt_licht_aus_an[]        = "Licht Aus/An    ";
static const __flash char txt_beleuchtung[]         = "Beleuchtung     ";
static const __flash char txt_musik_abspielen[]     = "Musik abspielen ";
static const __flash char txt_sleeptimer[]          = "Sleeptimer      ";
static const __flash char txt_musikstueck[]         = "Musikst"UML_u"ck      ";
static const __flash char txt_aufstehen[]           = "Aufstehen!      ";
static const __flash char txt_jetzt_komm_mal[]      = "Jetzt  komm  mal";
static const __flash char txt_aus_den_federn[]      = "aus den Federn !";
static const __flash char txt_alarmoptionen[]       = "Alarmoptionen   ";
static const __flash char txt_fade_in_dauer[]       = "Fade-In-Dauer   ";
static const __flash char txt_endlautstaerke[]      = "Endlautst"UML_a"rke   ";
static const __flash char txt_aufblenddauer[]       = "Aufblenddauer   ";
static const __flash char txt_zurueck_aendern_vor[] = "<<   "UML_a"ndern   >>";
static const __flash char txt_minminus_ok_minplus[] = "Min-   OK   Min+";
static const __flash char txt_stdplus_ok_minplus[]  = "Std+   OK   Min+";
static const __flash char txt_volminus_ok_volplus[] = "Vol-   OK   Vol+";
static const __flash char txt_nrminus_ok_nrplus[]   = "Nr.-   OK   Nr.+";
static const __flash char txt_volminus_volplus[] =    "Vol-        Vol+";
/***********************************************************************************************************************
* private Funktion
* erzeugt Enable Impuls
***********************************************************************************************************************/
static void display_enable (void)
{
  DISP_SET_E_HIGH;
  _delay_us(5);
  DISP_SET_E_LOW;
}
/***********************************************************************************************************************
* private Funktion
* legt 4 Bits an Display Datenleitungen D7-D4 --> PC3-PC0
* Übergabe des Nibbels im High-Nible von data
***********************************************************************************************************************/
static void display_nibble_out (uint8_t data)
{
  if ((data & 0b10000000) == 0) DISP_SET_D7_LOW; else DISP_SET_D7_HIGH;
  if ((data & 0b01000000) == 0) DISP_SET_D6_LOW; else DISP_SET_D6_HIGH;
  if ((data & 0b00100000) == 0) DISP_SET_D5_LOW; else DISP_SET_D5_HIGH;
  if ((data & 0b00010000) == 0) DISP_SET_D4_LOW; else DISP_SET_D4_HIGH;
  display_enable();
}
/***********************************************************************************************************************
* private Funktion
* sendet Textbyte an Display
***********************************************************************************************************************/
static void display_send_data_byte (uint8_t data)
{
  DISP_SET_RS_HIGH;
  display_nibble_out(data);
  display_nibble_out(data<<4);
}
/***********************************************************************************************************************
* private Funktion
* sendet Befehlsbyte an Display
***********************************************************************************************************************/
static void display_send_command_byte (uint8_t data)
{
  DISP_SET_RS_LOW;
  display_nibble_out(data);
  display_nibble_out(data<<4);
}
/***********************************************************************************************************************
* Timer0OverflowIRQ
*
* tritt alle 277,7µs auf (genug Verarbeitungszeit für jeden Befehl ans Display)
* Display komplett aufgebaut alle 35 IRQs <- Refresh Rate ca. 100 Hz
* Bei IRQ 0 und 17 wird lediglich die GCRAM-Adresse gesetzt um zu Beginn Zeile1 oder 2 zu springen
* bei IRQ 1-16 und 18-33 wird je ein Display-Zeichen ausgegeben
***********************************************************************************************************************/
ISR(TIMER0_OVF_vect)
{
  static uint8_t cnt=0;
  uint8_t chr;

  // IRQ-Zähler 0...33
  if (cnt >= 34) cnt=0; else cnt++;

  // IRQ = 0 Display Cursor auf Position Zeile1 setzen
  if (cnt == 0) display_send_command_byte(LCD_SET_DDADR + LCD_DDADR_LINE1);

  // IRQ 1...16 - ein Display-Zeichen in Zeile 1 aktualisieren
  else if ((cnt >= 1) && (cnt <= 16)) {
    chr = Display_Data[cnt - 1];
    display_send_data_byte(chr);
  }

  // IRQ = 17 Display Cursor auf Position Zeile2 setzen
  else if (cnt == 17) display_send_command_byte(LCD_SET_DDADR + LCD_DDADR_LINE2);

  // IRQ 18...33 - ein Display-Zeichen in Zeile 2 aktualisieren
  else if ((cnt >= 18) && (cnt <= 33)) {
    chr = Display_Data[cnt - 2];
    display_send_data_byte(chr);
  }
}
/***********************************************************************************************************************
* private Funktion
*
* kopiert Textbaustein aus Flash in Display Variable zur Anzeige
* Übergabe: Textbaustein
*           Startposition im Display 0...31
***********************************************************************************************************************/
static void cpy_flashtext_to_display (const char __flash *text, uint8_t startpos)
{
  while (*text) {
    Display_Data[startpos++] = *text;
    text++;
  }
}
/***********************************************************************************************************************
* private Funktion
*
* gibt Ziffer zweistellig auf Display aus
* Übergabe: Ziffer binär (0...99)
*           Ob führende Null an 10er-Stelle angezeigt werden soll (1 = ja, 0 = nein)
*           Startposition im Display 0...31
***********************************************************************************************************************/
static void cpy_number_to_display (uint8_t number, uint8_t zero, uint8_t pos)
{
  uint8_t tens = '0';
  uint8_t ones = '0';

  while (number >= 10) {
    tens++;
    number -= 10;
  }
  ones += number;

  if ((zero == 0) && (tens == '0')) tens = ' ';

  Display_Data[pos] = tens;
  Display_Data[pos+1] = ones;
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* Initialisiert Display und zeigt Begrüßung an
***********************************************************************************************************************/
void display_init (void)
{
  PORTC &= 0b11110000;                                      // D4, D5, D6, D7 = Low
  DDRC |=  0b00001111;                                      // D4, D5, D6, D7 = Ausgänge

  PORTB &= 0b11111010;                                      // E, RS = Low
  DDRB |=  0b00000101;                                      // E, RS = Ausgänge

  // Display Init Zeiten nach Datenblatt
  _delay_ms(15);
  display_nibble_out(LCD_SOFT_RESET);
  _delay_ms(5);
  display_enable();
  _delay_ms(1);
  display_enable();
  _delay_ms(1);
  display_nibble_out(LCD_SET_FUNCTION | LCD_FUNCTION_4BIT);
  _delay_ms(5);
  display_send_command_byte(LCD_SET_FUNCTION | LCD_FUNCTION_4BIT | LCD_FUNCTION_2LINE | LCD_FUNCTION_5X7);
  _delay_us(50);
  display_send_command_byte(LCD_SET_DISPLAY | LCD_DISPLAY_ON | LCD_CURSOR_OFF | LCD_BLINKING_OFF);
  _delay_us(50);
  display_send_command_byte(LCD_SET_ENTRY | LCD_ENTRY_INCREASE | LCD_ENTRY_NOSHIFT);
  _delay_us(50);
  display_send_command_byte (LCD_CLEAR_DISPLAY);
  _delay_ms (2);

  // Display-Refresh via IRQ
  TCCR0 = 0b00000010;                                       // Prescaler /8 -> Timer0OverflowIRQ mit 3600Hz
  TIMSK |= (1<<TOIE0);                                      // Timer0OverflowIRQ ein
  sei();
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* Ausgabe von Uhrzeit und Weckzeit auf Display <-- wenn KEIN Sleep-Timer aktiv
* ------------------     ------------------
* |Uhrzeit    12:00|     |Uhrzeit    12:00|
* |Alarmzeit  06:00|     |Alarmzeit    Aus|
* ------------------     ------------------
*  Alamschalter=Ein       Alarmschalter=Aus
*
* ------------------     ------------------
* |Uhrzeit[10]12:00|     |Uhrzeit[10]12:00|
* |Alarmzeit  06:00|     |Alarmzeit    Aus|
* ------------------     ------------------
* Ausgabe von Uhrzeit und Weckzeit auf Display <-- wenn Sleep-Timer aktiviert ist
* incl. noch aktivem Slepp-Timer in Minuten
************************************************************************************************************************/
void display_time_alarmtime (timehms_t cur_time, timehms_t alarm_time)
{
  uint8_t minutes_until_sleep = sleeptimer_get_minutes_until_sleep();

  cpy_flashtext_to_display(txt_uhrzeit, 0);                 // Text Uhrzeit
  cpy_number_to_display(cur_time.h, 0, 11);
  Display_Data[13] = ':';
  cpy_number_to_display(cur_time.m, 1, 14);

  if (minutes_until_sleep > 0) {                            // ggf Sleep-Minuten
    minutes_until_sleep--;
    Display_Data[7] = '[';
    cpy_number_to_display(minutes_until_sleep, 1, 8);
    Display_Data[10] = ']';
  }

  cpy_flashtext_to_display(txt_alarmzeit, 16);              // Alarmzeit
  if (alarm_switch_get_state() == SWITCH_ON) {
    cpy_number_to_display(alarm_time.h, 0, 27);
    Display_Data[29] = ':';
    cpy_number_to_display(alarm_time.m, 1, 30);
  }
  else {                                                    // Alarmzeit = Aus
    Display_Data[29] = 'A';
    Display_Data[30] = 'u';
    Display_Data[31] = 's';
  }
}
/***********************************************************************************************************************
* öffentliche Funktion zur Ausgabe von:
*
* ------------------
* |ROOT Tracks   22|
* |Spiele  Nr.1  an|
* ------------------
* Übergeben werden muß die Anzahl der Tracks im Root Verzeichnis
***********************************************************************************************************************/
void display_tracks_in_root (uint8_t tracks)
{
  cpy_flashtext_to_display(txt_tracks_in_root, 0);
  cpy_number_to_display(tracks, 0, 14);
  cpy_flashtext_to_display(txt_spiele_nr1_an, 16);
}
/***********************************************************************************************************************
* Private Funktion
*
* Zeigt Alarmzeit an und läßt diese ändern
* ------------------
* |Alarmzeit  06:00|
* |Std+   OK   Min+|
* ------------------
***********************************************************************************************************************/
static void mnchange_alarmtime (void)
{
  uint8_t btn;
  uint8_t alarm_h = eemem_read_data(EEDATA_ALARM_STUNDE);
  uint8_t alarm_m = eemem_read_data(EEDATA_ALARM_MINUTE);

ShowValue:
  cpy_flashtext_to_display(txt_alarmzeit, 0);
  cpy_number_to_display(alarm_h, 0, 11);
  Display_Data[13] = ':';
  cpy_number_to_display(alarm_m, 1, 14);
  cpy_flashtext_to_display(txt_stdplus_ok_minplus, 16);
ReadButtons:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    alarm_h++;
    if (alarm_h > 23) alarm_h = 0;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    alarm_m++;
    if (alarm_m > 59) alarm_m = 0;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_MID) {
    eemem_write_data(EEDATA_ALARM_STUNDE, alarm_h);
    eemem_write_data(EEDATA_ALARM_MINUTE, alarm_m);
    buttons_wait_release();
    return;
  }
  goto ReadButtons;
}
/***********************************************************************************************************************
* Private Funktion
*
* Zeigt Uhrzeit an und läßt diese ändern
* ------------------
* |Uhrzeit    12:00|
* |Std+   OK   Min+|
* ------------------
***********************************************************************************************************************/
static void mn_change_time (void)
{
  uint8_t btn;
  timehms_t curtime = rtc_get_time();

ShowValue:
  cpy_flashtext_to_display(txt_uhrzeit, 0);
  cpy_number_to_display(curtime.h, 0, 11);
  Display_Data[13] = ':';
  cpy_number_to_display(curtime.m, 1, 14);
  cpy_flashtext_to_display(txt_stdplus_ok_minplus, 16);
ReadButtons:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    curtime.h++;
    if (curtime.h > 23) curtime.h = 0;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    curtime.m++;
    if (curtime.m > 59) curtime.m = 0;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_MID) {
    curtime.s = 0;
    rtc_set_time(curtime);
    buttons_wait_release();
    return;
  }
  goto ReadButtons;
}
/***********************************************************************************************************************
* private Funktion
*
* läßt die Alarmoptionen ändern
* ------------------
* |Aufblenddauer 10|
* |Min-   OK   Min+|
* ------------------
* |Fade-In-Dauer 20|
* |Min-   OK   Min+|
* ------------------
* |Endlautstärke 30|
* |Vol-   OK   Vol+|
* ------------------
***********************************************************************************************************************/
static void mn_change_alarmoptions (void)
{
  uint8_t btn;
  uint8_t light = eemem_read_data(EEDATA_AUFBLENDDAUER);
  uint8_t fade  = eemem_read_data(EEDATA_FADEIN_DAUER);
  uint8_t vol   = eemem_read_data(EEDATA_ENDVOLUME);

ShowLight:
  cpy_flashtext_to_display(txt_aufblenddauer, 0);
  cpy_number_to_display(light, 0, 14);
  cpy_flashtext_to_display(txt_minminus_ok_minplus, 16);
ReadBTNLight:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    light--;
    if (light < 10) light = 30;
    goto ShowLight;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    light++;
    if (light > 30) light = 10;
    goto ShowLight;
  }
  else if (btn == BUTTON_PRESSED_NONE) goto ReadBTNLight;

ShowFade:
  cpy_flashtext_to_display(txt_fade_in_dauer, 0);
  cpy_number_to_display(fade, 0, 14);
  cpy_flashtext_to_display(txt_minminus_ok_minplus, 16);
ReadBTNFade:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    fade--;
    if (fade < 10) fade = 30;
    goto ShowFade;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    fade++;
    if (fade > 30) fade = 10;
    goto ShowFade;
  }
  else if (btn == BUTTON_PRESSED_NONE) goto ReadBTNFade;

ShowVol:
  cpy_flashtext_to_display(txt_endlautstaerke, 0);
  cpy_number_to_display(vol, 0, 14);
  cpy_flashtext_to_display(txt_volminus_ok_volplus, 16);
ReadBTNVol:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    vol--;
    if (vol < 10) vol = 30;
    goto ShowVol;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    vol++;
    if (vol > 30) vol = 10;
    goto ShowVol;
  }
  else if (btn == BUTTON_PRESSED_NONE) goto ReadBTNVol;

  eemem_write_data(EEDATA_AUFBLENDDAUER, light);
  eemem_write_data(EEDATA_FADEIN_DAUER, fade);
  eemem_write_data(EEDATA_ENDVOLUME, vol);
  buttons_wait_release();
}
/***********************************************************************************************************************
* Private Funktion
*
* Schaltet das Licht manuell in Stufen von 0...25 --> Manuell an/aus
* ------------------
* |Beleuchtung   10|
* |Nr-    OK    Nr+|
* ------------------
***********************************************************************************************************************/
void mn_change_light (void)
{
  uint8_t btn;
  uint8_t light = 0;

ShowValue:
  cpy_flashtext_to_display(txt_beleuchtung, 0);
  cpy_number_to_display(light, 0, 14);
  cpy_flashtext_to_display(txt_nrminus_ok_nrplus, 16);
  light_set(light * 10);
ReadButtons:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    if (light > 0) light--;
    else light = 25;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    if (light < 25) light++;
    else light = 0;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_MID) {
    buttons_wait_release();
    return;
  }
  goto ReadButtons;
}
/***********************************************************************************************************************
* Private Funktion
*
* Läßt eine Schlummermusik wählen bzw. startet / beendet Musikausgabe
* Anzeige/Auswahl Musikstücknummer 2...Anzahl auf Karte bzw. Aus und Ausgabe stoppen
* ------------------
* |Musikstück    99|
* |Nr-    OK    Nr+|
* ------------------
***********************************************************************************************************************/
void mn_change_music (void)
{
  uint8_t no = 0;
  uint8_t tracks;
  uint8_t btn;

  cpy_flashtext_to_display(txt_moment, 0);
  cpy_flashtext_to_display(txt_starte_mp3modul, 16);
  mp3_player_wakeup();
  tracks = mp3_player_get_tracks();

ShowValue:
  cpy_flashtext_to_display(txt_musikstueck, 0);
  cpy_flashtext_to_display(txt_nrminus_ok_nrplus, 16);
  if (no == 0) {
    Display_Data[13] = 'A';
    Display_Data[14] = 'u';
    Display_Data[15] = 's';
  }
  else cpy_number_to_display(no, 0, 14);
ReadButtons:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    if (no == 2) no = 0;
    else if (no == 0) no = tracks;
    else no--;
    mp3_player_set_volume(20);
    if (no != 0) mp3_player_play(no);
    else mp3_player_stop();
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    if (no == tracks) no = 0;
    else if (no == 0) no = 2;
    else no++;
    mp3_player_set_volume(20);
    if (no != 0) mp3_player_play(no);
    else mp3_player_stop();
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_MID) {
    if (no == 0) {
      mp3_player_stop();
      mp3_player_sleep();
    }
    buttons_wait_release();
    return;
  }
  goto ReadButtons;
}
/***********************************************************************************************************************
* Private Funktion
*
* Läßt Sleep-Timer setzten
* ------------------
* |Sleeptimer    10| Anzeige Sleep-Timer-Zeit in Minuten (10...60) oder Aus
* |Min-   OK   Min+|
* ------------------
***********************************************************************************************************************/
void mn_change_sleep (void)
{
  uint8_t min = 0;
  uint8_t btn;

ShowValue:
  cpy_flashtext_to_display(txt_sleeptimer, 0);
  cpy_flashtext_to_display(txt_minminus_ok_minplus, 16);
  if (min == 0) {
    Display_Data[13] = 'A';
    Display_Data[14] = 'u';
    Display_Data[15] = 's';
  }
  else cpy_number_to_display(min, 0, 14);
ReadButtons:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    if (min == 0) min = 60;
    else if (min == 10) min = 0;
    else min--;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    if (min == 60) min = 0;
    else if (min == 0) min = 10;
    else min++;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_MID) {
    if (min != 0) min +=2;
    sleeptimer_set_minutes_until_sleep(min);
    buttons_wait_release();
    return;
  }
  goto ReadButtons;
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* zeigt Hauptmenu mit Punkten Ebene1 an, läßt auswählen und springt die jeweilige Unterroutine zur Änderung an
* kehrt zurück sobald die Einstellungen getätigt wurden
* oder in Menüebene 1 3 Sekunden lang kein Tasterdruck erfolgt
***********************************************************************************************************************/
void display_menu (void)
{
  uint8_t mn = 0;
  uint8_t btn;

ShowValue:
  if (mn == 0) cpy_flashtext_to_display(txt_alarmzeit, 0);
  else if (mn == 1) cpy_flashtext_to_display(txt_sleeptimer, 0);
  else if (mn == 2) cpy_flashtext_to_display(txt_licht_aus_an, 0);
  else if (mn == 3) cpy_flashtext_to_display(txt_musik_abspielen, 0);
  else if (mn == 4) cpy_flashtext_to_display(txt_alarmoptionen, 0);
  else cpy_flashtext_to_display(txt_uhrzeit, 0);
  cpy_flashtext_to_display(txt_zurueck_aendern_vor, 16);
  ticker100ms_start(TICKER_DISPLAY);

ReadButtons:
  buttons_wait_release();
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_LEFT) {
    if (mn > 0) mn--; else mn = 5;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {
    if (mn < 5) mn++; else mn = 0;
    goto ShowValue;
  }
  else if (btn == BUTTON_PRESSED_MID) {
    if (mn == 0) {
      mnchange_alarmtime();
      return;
    }
    else if (mn == 1) {
      mn_change_sleep();
      return;
    }
    else if (mn == 2) {
      mn_change_light();
      return;
    }
    else if (mn == 3) {
      mn_change_music();
      return;
    }
    else if (mn == 4) {
      mn_change_alarmoptions();
      return;
    }
    else if (mn == 5) {
      mn_change_time();
      return;
    }
  }
  if (ticker100ms_get_value(TICKER_DISPLAY) < 30) goto ReadButtons;
}
/***********************************************************************************************************************
* öffentliche Funktion --> zeigt:
* ------------------      ------------------
* |Aufstehen! 06:10|      |Aufstehen! 06:10|
* |Vol-  ^10^  Vol+|      |Vol-  <10>  Vol+|
* ------------------      ------------------
*  Wenn val_fadeup != 0    Wenn vol_fadeup == 0
***********************************************************************************************************************/
void display_wakeup (timehms_t cur_time, uint8_t cur_vol, uint8_t vol_fadeup)
{
  cpy_flashtext_to_display(txt_aufstehen, 0);
  cpy_number_to_display(cur_time.h, 0, 11);
  Display_Data[13] = ':';
  cpy_number_to_display(cur_time.m, 1, 14);
  cpy_flashtext_to_display(txt_volminus_volplus, 16);
  cpy_number_to_display(cur_vol, 1, 23);
  if (vol_fadeup == 0) {
    Display_Data[22] = '<';
    Display_Data[25] = '>';
  }
  else {
    Display_Data[22] = '^';
    Display_Data[25] = '^';
  }
}
/***********************************************************************************************************************
* öffentliche Funktion --> zeigt:
* ------------------
* |Jetzt  komm  mal|
* |aus den Federn !|
* ------------------
***********************************************************************************************************************/
void display_alarmtext (void)
{
  cpy_flashtext_to_display(txt_jetzt_komm_mal, 0);
  cpy_flashtext_to_display(txt_aus_den_federn, 16);
}
