/*
 * light.c
 *
 * Copyright 2019 Thomas Gollmer <th_goso@freenet.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 * Ansteuerung Licht über OC1A -> PB1
 * Weiterhin werden via IRQ Timer/Ticker realisiert zum zählen der Zeiten
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include "light.h"

// die Tickervariablen <-- alle 100ms basierend
static volatile uint16_t Ticker_100ms[TICKERS];
static volatile uint16_t Sleep_Timer = 0;
static volatile uint16_t BL_Timer = 0;


/***********************************************************************************************************************
 * Timer2OverflowIRQ
 *
 * tritt alle 2,222222 ms auf (nach 45 IRQs sind 100ms vergangen)
 * es werden alle Timer/Ticker jede Sekunde raufgezählt (kein Überlauf, Ende bei FFFF)
 * Der Sleep-Timer wird alle 100ms heruntergezählt (kein Unterlauf, sondern Ende bei 0)
 * Der BackLightTimer wird alle 100ms heruntergezählt (kein Unterlauf, sondern Ende bei 0)
 *                    wenn bei 1 angekommen, wird die Hintergrundbeleuchtung abgeschaltet
 **********************************************************************************************************************/
ISR(TIMER2_OVF_vect)
{
  static uint8_t irqcnt = 0;
  uint8_t tickercnt;

  if (irqcnt < 45) {
    irqcnt++;
    return;
  }

  else {
    irqcnt = 0;
    for (tickercnt=0; tickercnt<TICKERS; tickercnt++) if (Ticker_100ms[tickercnt] < 0xFFFF) Ticker_100ms[tickercnt]++;
    if (Sleep_Timer != 0) Sleep_Timer--;
    if (BL_Timer != 0) BL_Timer--;
    if (BL_Timer == 1) PORTD &= ~(1<<PD6);
  }
}
/***********************************************************************************************************************
 * öffentliche Funktion
 *
 * Initalisiert Timer1 für LED-PWM und Timer2 für Ticker
 **********************************************************************************************************************/
void light_init (void)
{
  PORTB |= (1<<PB1);                                        // Licht aus
  DDRB |= (1<<PB1);                                         // PB1 = Ausgang LED Beleuchtung
  PORTD |= (1<<PD6);                                        // Licht an
  DDRD |= (1<<PD6);                                         // PD6 = Ausgang Displaybeleuchtung

  OCR1A = 0;                                                // Licht aus
  TCCR1A = 0b11000001;                                      // Fast PWM 8 Bit Inverted
  TCCR1B = 0b00001001;                                      // precaler = /1 --> 28,8 kHz
  TCCR2 = 0b00000100;                                       // Timer2 prescaler = /64 -> Overflow alle 2,222ms
  TIMSK |= (1<<TOIE2);                                      // Timer2OverflowIRQ ein
  sei();
}
/***********************************************************************************************************************
 * öffentliche Funktion
 *
 * setzt die LED Beleuchtung auf übergebene Helligkeitsstufe 0...255
 * ggf kann hier eine PWM-Tabelle eingebaut werden, jedoch sieht es mit
 * linearem Anstieg der Helligkeit am besten aus um aufzuwachen
 **********************************************************************************************************************/
void light_set (uint8_t value)
{
  OCR1A= value;
}
/***********************************************************************************************************************
 * öffentliche Funktion
 *
 * resetet/startet einen Ticker
 * Übergabe: Ticker-Nummer
 **********************************************************************************************************************/
void ticker100ms_start (uint8_t ticker_no)
{
  if (ticker_no >= TICKERS) return;

  cli();
  Ticker_100ms[ticker_no] = 0;
  sei();
}
/***********************************************************************************************************************
 * öffentliche Funktion
 *
 * liefert den aktuellen Stand eines Count-Up-Timers zurück
 * Übergabe: Ticker-Nummer
 **********************************************************************************************************************/
uint16_t ticker100ms_get_value (uint8_t ticker_no)
{
  uint16_t ret;

  if (ticker_no >= TICKERS) return 0xFFFF;
  cli();
  ret = Ticker_100ms[ticker_no];
  sei();
  return ret;
}
/***********************************************************************************************************************
 * öffentliche Funktion
 *
 * Startet den Sleep-Timer (Count-Down-Timer)
 * Übergabe: Zeit bis zum Timeout in Minuten
 **********************************************************************************************************************/
void sleeptimer_set_minutes_until_sleep (uint8_t minutes)
{
  uint16_t tmp = minutes;                                   // Minuten nach
  tmp *= 600;                                               // 100er Millisekunden umrechnen
  cli();                                                    // 100ms basierenden Sleep-Timer starten
  Sleep_Timer = tmp;                                        // wird jetzt im IRQ alle 100ms um 1 dekrementiert
  sei();
}
/***********************************************************************************************************************
 * öffentliche Funktion
 * Holt den aktuellen Stand des Sleep-Timers ab
*
 * Rückgabe: aktuelle Zeit bis zum Timeout in Minuten
 **********************************************************************************************************************/
uint8_t sleeptimer_get_minutes_until_sleep (void)
{
  uint16_t tmp;

  cli();
  tmp = Sleep_Timer;
  sei();
  tmp /= 600;                                               // Sleep-Timer 100ms basierend
  return (uint8_t)(tmp & 0xFF);                             // umrechen in Minuten und zurückgeben
}
/***********************************************************************************************************************
 * öffentliche Funktion
 * Schaltet die Displayhintergrundbeleuchtung an PD6 ein
 * via IRQ wird die Beleuchtung nach der übergebenern Dauer wieder abgeschaltet
 *
 * Übergabe: Ein-Zeit in Sekunden
 **********************************************************************************************************************/
void backlight_set_on (uint8_t seconds)
{
  uint16_t tmp = seconds;

  PORTD |= (1<<PD6);                                        // Displaybeleuchtung einschalten
  tmp *= 10;                                                // Sekunden nach 100ms umrechnen
  cli();
  BL_Timer = tmp;                                           // BackLight-Timer starten (zählt im IRQ herunter)
  sei();
}
