/*
 * Lichtwecker mit RTC-Modul DS3231 und MP3-Modul MP3-TF-16 sowie HD44780 kompatiblem Display 2 Zeilen je 16 Zeichen

 * Copyright 2019 Thomas Gollmer <th_goso@freenet.de>

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.

 * CPU:                ATMega8
 * Clock:              7,3728 MHz Quarz
 * Interne Hardware:   Timer0OverflowIRQ zum Displayrefresh ca. 112Hz
 *                     Timer2OverflowIRQ für Ticker/Timer als Zeitbasis
 *                     Timer1 incl. OC1A zur PWM Erzeugung für LED Beleuchtung (erhöht auf ca. 29 kHz)
 *                                           Alles niedrigere, machte sich in den Lautsprechern bemerkbar
 *
 * Anschlüsse:
 *                                                                                PD6 - Ausgang LED+ zum Display
 * PB5 - ISP                                  PC5 - Ausgang I2C = SCL             PD5 - Eingang = Schalter Alarm An/Aus
 * PB4 - ISP                                  PC4 - Ein/Ausgang I2C = SDA         PD4 - Eingang = Taste Rechts
 * PB3 - ISP                                  PC3 - Ausgang = Display D7          PD3 - Eingang = Taste Mitte
 * PB2 - Ausgang = Display RS                 PC2 - Ausgang = Display D6          PD2 - Eingang = Taste Links
 * PB1 - Ausgang OC1A = LED Beleuchtung       PC1 - Ausgang = Display D5          PD1 - Ausgang UART-TX = MP3-Modul
 * PB0 - Ausgang = Display Enable             PC0 - Ausgang = Display D4          PD0 - Eingang UART-RX = MP3-Modul
 *
 *
 *
 * Normalbetrieb - Displaymöglichkeiten bei normalem Betrieb/Zeitanzeige
 * ----------------------------------- Linke/Rechte Taste ändert Laustärke eines ggf. laufenden "Schlummerstückes"
 * |Uhrzeit    12:00|Uhrzeit    12:00| Mit Druck auf mittlere Taste wird das Auswahlmenü angezeigt
 * |Alarmzeit  06:00|Alarmzeit    Aus| Falls dann 3 Sekunden nichts gewählt/geändert wird,
 * ----------------------------------- wechselt die Anzeige wieder in diesen Modus hier
 * |Uhrzeit[10]12:00|Uhrzeit[10]12:00|
 * |Alarmzeit  06:00|Alarmzeit    Aus| Anzeige wenn Sleep-Modus aktiv
 * ----------------------------------- neben der Uhrzeit werden noch die Aktivminuten ausgegeben
 *
 *
 * Alarmanzeige - Musik aufdrehen und Licht aufblenden:
 * ------------------  > Alarm beenden wenn Schalter Alarm An/Aus auf aus geschaltet wird
 * |Aufstehen! 06:15|  > Rechte oder Linke Taste stoppt Musik-Fade-In und stellt auf feste Lautstärke
 * |Vol-  ^15^  Vol+|  > Alarmmodus läuft jedoch weiter (Licht aufblenden und  Extraweckton abspielen)
 * ------------------
 *
 *
 * Anzeige wenn Musik aufdrehen/Licht aufblenden vergangen ist und nun der Extraweckton abgespielt wird (30 Sek lang)
 * ------------------
 * |Jetzt  komm  mal|  > Lediglich kippen von Schalter Alarm an/Aus auf aus beendet nun den Alarm
 * |aus den Federn !|  > Falls nach 30 Sekunden nun immer noch kein Alarm ausgeschaltet wurde, Alarm beenden
 * ------------------
 *
 * Menustruktur
 * Auswahlmenü - Menüebene 1
 * Punkte werden nacheinander angezeigt, falls 3 Sekunden kein Tastendruck zurück in Normalbetrieb
 *
 *                  für Sleeptimer --> Manuell Licht    Manuell Sound
 * -------------------------------------------------------------------------------------------------------
 * |Alarmzeit       |Sleeptimer      |Licht Aus/An    |Musik abspielen |Alarmoptionen   |Uhrzeit         |
 * |<<   ändern   >>|<<   ändern   >>|<<   ändern   >>|<<   ändern   >>|<<   ändern   >>|<<   ändern   >>|
 * -------------------------------------------------------------------------------------------------------
 * |Alarmzeit  06:00|Sleeptimer    30|Beleuchtung   20|Moment bitte !  |Aufblenddauer 10|Uhrzeit    12:00|
 * |Std+   OK   Min+|Min-   OK   Min+|Nr+    OK    Nr-|Starte MP3-Modul|Min-   OK   Min+|Std+   OK   Min+|
 * -------------------------------------------------------------------------------------------------------
 *                   Sleeptimer                       |Musikstück    99|Fade-In-Dauer 20|
 *                   Zeit bis zum                     |Nr-    OK    Nr+|Min-   OK   Min+|
 *                   Ausschalten von                  -----------------------------------
 *                   Manuellem                                         |Endlautstärke 30|
 *                   Musikstück/Licht                                  |Vol-   OK   Vol+|
 *                   10-60 Minuten                                     ------------------
 *                                                   Musikstück         Aufblenddauer Licht 10-30 Minuten
 *                                                   2...Anzahl Karte   Fade-In-Dauer Musik 10-30 Minuten
 *                                                   oder Aus           Endlautstärke Stufe 10-30
 *
*/

#include <stdlib.h>
#include <avr/io.h>
#include <util/delay.h>
#include "rtc.h"
#include "display.h"
#include "mp3player.h"
#include "buttons.h"
#include "eemem.h"
#include "light.h"

/***********************************************************************************************************************
 * Startinitialisierungen, I/Os Ports, Hardware
 * Dauer ca. 10 Sekunden mit Displayanzeigen und Infoausgaben
 * Gibt die Anzahl der auf der SD-Karte im root enthaltenen Musikstücke zurück
 **********************************************************************************************************************/
static uint8_t startinit (void)
{
  uint8_t tracks;
  uint8_t light;
  timehms_t curtime;

  light_init();                                             // LED/Timer1 init
  rtc_init();                                               // DS3231/I2C-Bus initialisieren
  buttons_init();                                           // Schalter, Tastereingänge
  mp3_player_init();                                        // FNM16P-MP3-Modul/UART init
  display_init();                                           // Displayinit & kurze Textanzeige 5 Sekunden warten
  _delay_ms(5000);                                          // Damit das MP3-Modul durchstarten kann
  tracks = mp3_player_get_tracks();                         // Anzahl Musikstücke auf SD-Karte lesen
  display_tracks_in_root(tracks);                           // Anzahl speichern und ausgeben
  mp3_player_set_volume(20);                                // nicht ganz sooo laut
  mp3_player_play(1);                                       // Extraweckton kurz anspielen

  for (light=0; light<255; light++) {                       // 5 Sekunden Licht
    light_set(light);                                       // Auf- und wieder Abblenden
    _delay_ms(10);                                          // So lange spielt auch der Weckton
  }
  for (light=255; light>0; light--) {
    light_set(light);
    _delay_ms(10);
  }
  light_set(0);

  mp3_player_stop();                                        // Stoppen und
  mp3_player_sleep();                                       // Modul schlafen legen
  curtime = rtc_get_time();                                 // Bei start aktuelle Zeit holen
  srand(curtime.h + curtime.m + curtime.s);                 // und als Seed für Zufallsgenerator nehmen
  return tracks;
}
/***********************************************************************************************************************
 * Alarmroutine -> Anspringen wenn die Alarmzeit erreicht ist und Alarmschalter auf ON steht
 *
 * Routine kehrt zurück wenn:
 * - Alarm per Wahlschalter Alarm An/Aus beendet wurde
 * - Fade-In-Zeit, Light-In-Zeit verstrichen und zusätzlich als letztes der Extraweckton 30 Sekunden abgespielt
 *
 * Vor Rückkehr werden Musikausgabe und Licht wieder abgeschaltet
 **********************************************************************************************************************/
static void alarm (uint8_t tracks, uint8_t fade_in_time, uint8_t end_vol, uint8_t light_in_time)
{
  timehms_t cur_time = rtc_get_time();
  uint16_t  rev_light_time;
  uint16_t  rev_fade_time;
  uint8_t   rnd_no;
  uint8_t   btn;
  uint8_t   stop_vol=0;
  uint8_t   cur_vol=0;
  uint8_t   cur_light=1;

  rev_light_time = light_in_time;                           // light_in_time           (10...30) Minuten
  rev_light_time *= 600;                                    // Minuten nach 100ms      (6000...18000)
  rev_light_time /= 256;                                    // durch PWM Stufen teilen (23...70)
  // 23...70 mal 100ms jede PWM Stufen durchlaufen um auf fade_in_time zu kommen

  rev_fade_time = fade_in_time;                             // fade_in_time                 (10...30) Minuten
  rev_fade_time *= 600;                                     // Minuten nach 100ms           (6000...18000)
  rev_fade_time /= end_vol;                                 // durch Endvol(10...30) teilen (200...1800)
  // 200...1800 mal 100ms jede Volume-Stufe durchlaufen um auf fade_in_time zu kommen

  tracks--;                                                 // Zufällige Tracknummer
  rnd_no = rand();                                          // aus Bereich
  rnd_no = rnd_no % tracks;                                 // 2...tracks
  rnd_no += 2;                                              // erzeugen
  if (rnd_no < 2) rnd_no = 2;                               // Sicherheitshalber
  if (rnd_no > tracks) rnd_no = tracks;                     // sollte aber nicht vorkommen

  display_wakeup(cur_time, cur_vol, 1);                     // Alarmtext 1 anzeigen
  light_set(cur_light);                                     // Beleuchtung niedrigste Stufe
  mp3_player_wakeup();
  mp3_player_set_volume(cur_vol);                           // Sound niedrigste Stufe
  mp3_player_play(rnd_no);                                  // zufälliges Stück spielen
  ticker100ms_start(TICKER_LIGHTIN);                        // Die 3 Ticker starten
  ticker100ms_start(TICKER_FADEIN);;                        // Für Fade-In, Light-In
  ticker100ms_start(TICKER_DISPLAY);                        // und Display -> für Zeitaktualisierung

  // Prüfen ob Aufbelndzeit, Fade-In-Zeit, Zeit zur Anzeigeaktualisierung um, je nach dem Aktion durchführen
  // Sprung zu Alarm End, wenn Alarm deaktiviert wird
  while (1) {
    // Display aktualisieren alle 10 Sekunden mit neuer Zeit
    if (ticker100ms_get_value(TICKER_DISPLAY) >= 100) {
      cur_time = rtc_get_time();
      ticker100ms_start(TICKER_DISPLAY);
      if (stop_vol == 0) display_wakeup(cur_time, cur_vol, 1);
      else display_wakeup(cur_time, stop_vol, 0);
    }
    // Light-In-Zeit aufblenden
    if (ticker100ms_get_value(TICKER_LIGHTIN) >= rev_light_time) {
      if (cur_light < 255) {
        cur_light++;
        light_set(cur_light);
      }
      ticker100ms_start(TICKER_LIGHTIN);
    }
    // Fade-In-Zeit Sound hochdrehen, aber nur, wenn Lautstärke NICHT per Knopfdruck festgesetzt wurde
    if (ticker100ms_get_value(TICKER_FADEIN) >= rev_fade_time) {
      if (cur_vol < end_vol) {
        cur_vol++;
        if (stop_vol == 0) {
          mp3_player_set_volume(cur_vol);
          display_wakeup(cur_time, cur_vol, 1);
        }
        else {
          mp3_player_set_volume(stop_vol);
          display_wakeup(cur_time, stop_vol, 0);
        }
      }
      ticker100ms_start(TICKER_FADEIN);
    }

    if ((cur_light == 255) && (cur_vol >= end_vol)) break;  // Wenn Zeit um ohne daß Alarm ausgeschaltet wurde
    if (alarm_switch_get_state()==SWITCH_OFF)goto AlarmEnd; // Alarm beenden wenn Schalter auf Aus geschaltet wurde
    btn = buttons_get_state();                              // Tasten lesen

    if (btn == BUTTON_PRESSED_LEFT) {                       // Linke Taste verringert Lautstärke
      if (stop_vol == 0) stop_vol = cur_vol;                // und setzt Lautstärke fest
      if (stop_vol > 1) stop_vol--;
      mp3_player_set_volume(stop_vol);
      display_wakeup(cur_time, stop_vol, 0);
      buttons_wait_release();
    }
    else if (btn == BUTTON_PRESSED_RIGHT) {                 // Rechte Taste erhöht Lautstärke
      if (stop_vol == 0) stop_vol = cur_vol;                // und setzt Lautstärke fest
      if (stop_vol < end_vol) stop_vol++;
      mp3_player_set_volume(stop_vol);
      display_wakeup(cur_time, stop_vol, 0);
      buttons_wait_release();
    }
  }

  // Zweiter Alarm... hier geht's weiter wenn Fade-In und Aufbelndzeit um sind
  // ohne das der Alarm ausgeschaltet wurde
  display_alarmtext();                                      // Alarmtext anzeigen
  mp3_player_set_volume(end_vol);                           // Endlautstärke setzten
  mp3_player_play(1);                                       // und den Extraweckton aus Musikdatei1 abspielen
  ticker100ms_start(TICKER_DISPLAY);                        // Timeout in 30 Sekunden
  while (ticker100ms_get_value(TICKER_DISPLAY) < 300) {
    if (alarm_switch_get_state()==SWITCH_OFF)goto AlarmEnd; // Alarm beenden wenn Schalter auf Aus gestellt wurde
  }

AlarmEnd:
  buttons_wait_release();                                   // Alarm beenden, egal von wo hier her gekommen
  mp3_player_stop();
  mp3_player_sleep();
  light_set(0);
}
/***********************************************************************************************************************
 * Hauptprogramm
 **********************************************************************************************************************/
int main (void)
{
  timehms_t alarm_time;                                     // Alarmzeit
  timehms_t cur_time;                                       // Aktuelle Uhrzeit
  uint8_t stat_switch = SWITCH_OFF;                         // Status Alarmschalter An/Aus
  uint8_t tracks;                                           // Anzahl Musikstücke im root der SD-Karte
  uint8_t fade_in_time;                                     // Einblenddauer in Minuten für Musikstücke
  uint8_t end_vol;                                          // Endlautstärke nach Faden-In
  uint8_t light_in_time;                                    // Ausfblenddauer Licht in Minuten
  uint8_t btn;                                              // Tasterstatus

  tracks = startinit();                                     // Init, Infoausgabe, Anzahl Musikstücke auf SD-Karte holen
LoadEEData:
  alarm_time.h = eemem_read_data(EEDATA_ALARM_STUNDE);      // Alarmzeit
  alarm_time.m = eemem_read_data(EEDATA_ALARM_MINUTE);      // und Fade-In-Dauer, Aufblenddauer, Endvolumen
  alarm_time.s = 0;                                         // aus EEPROM in lokale Variablen laden
  fade_in_time = eemem_read_data(EEDATA_FADEIN_DAUER);
  end_vol = eemem_read_data(EEDATA_ENDVOLUME);              // immer nach Neustart oder Rückkehr
  light_in_time = eemem_read_data(EEDATA_AUFBLENDDAUER);    // aus der Hauptmenüanzeige
ReadTime:
  ticker100ms_start(TICKER_DISPLAY);                        // 100ms basierenden Count-Up-Ticker starten
  cur_time = rtc_get_time();                                // aktuelle Uhrzeit abholen
  display_time_alarmtime(cur_time, alarm_time);             // Zeitanzeige aktualisieren

  if (sleeptimer_get_minutes_until_sleep() == 1) {          // Sleep-Timer war aktiv und ist nun abgelaufen
    mp3_player_stop();                                      // ggf. laufende Musikausgabe stoppen
    mp3_player_sleep();                                     // MP3 Modul schlafen legen
    light_set(0);                                           // Licht ausschalten
    sleeptimer_set_minutes_until_sleep(0);                  // Sleep-Timer abschalten
    goto ReadTime;                                          // Anzeige aktualisieren
  }

  // Wenn Alarmzeit erreicht, Alarmroutine anspringen
  if ((alarm_switch_get_state() == SWITCH_ON) && (alarm_time.h == cur_time.h) && (alarm_time.m == cur_time.m)) {
    sleeptimer_set_minutes_until_sleep(0);                  // ggf laufenden Sleep-Timer bei Alarm
    alarm (tracks, fade_in_time, end_vol, light_in_time);   // IMMER ausschalten
    goto ReadTime;                                          // Nach Rückkehr Zeit aktualisieren
  }

ReadButtons:
  btn = buttons_get_state();
  if (btn == BUTTON_PRESSED_MID) {                          // Falls Tastendruck mittlere Taste
    display_menu();                                         // Hauptmenü anzeigen und Änderungen tätigen lassen
    goto LoadEEData;                                        // geänderte Werte neu laden
  }
  else if (btn == BUTTON_PRESSED_LEFT) {                    // Linke Taste
    mp3_player_dec_volume();                                // ändert Lautstärke eines ggf. laufenden Stückes
    buttons_wait_release();
    goto ReadButtons;
  }
  else if (btn == BUTTON_PRESSED_RIGHT) {                   // Rechte Taste
    mp3_player_inc_volume();                                // ändert Lautstärke eines ggf. laufenden Stückes
    buttons_wait_release();
    goto ReadButtons;
  }
  if (alarm_switch_get_state() != stat_switch) {            // Änderungen an Alarm An/Aus Schalter
    stat_switch = alarm_switch_get_state();                 // Sofort anzeigen
    goto ReadTime;
  }
  if (ticker100ms_get_value(TICKER_DISPLAY) >= 100) {       // jeweils nach 10 Sekunden Zeit
    goto ReadTime;                                          // neu von RTC lesen
  }
  goto ReadButtons;                                         // Zwischendurch immer Buttons abfragen
}
