/*
 * mp3player.c
 *
 * Copyright 2019 Thomas Gollmer <th_goso@freenet.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 * MP3-Player Funktionen aufbauend auf Low-Level-Routinen für Modul fnm16P
 */

#include <avr/io.h>
#include <util/delay.h>
#include "mp3player.h"

// ---------------------------------------------------------------------------------------------------------------------
// Benötigte DS Register/Adressen für Funktionsübergaben
#define FNMCMD_SLEEP          0x0A      // Chip schlafen legen, Param = 0
#define FNMCMD_RESET          0x0C      // Player zurücksetzen, Param = 0

#define FNMCMD_SET_TRACK      0x03      // Track aus Wurzelverzeichnis anwählen, Param = Tracknummer 1...xxx
#define FNMCMD_INC_VOLUME     0x04      // Increase Volume, Param = 0
#define FNMCMD_DEC_VOLUME     0x05      // Decrease volume, Param = 0
#define FNMCMD_SET_VOLUME     0x06      // Ausgangslautstärke setzten, Param = 1...30
#define FNMCMD_SET_EQ         0x07      // EQ Mode setzen, Param = 0-Normal 1-Pop 2-Rock 3-Jazz 4-Classic 5-Bass
#define FNMCMD_SET_DEVICE     0x09      // Abspielgerät setzen, Param = 1 = USB, 2 = SD-Card

#define FNMCMD_PLAY           0x0D      // Play (vorher einen Track 'anwählen'), Param = 0
#define FNMCMD_STOP           0x16      // Stop, Param = 0

#define FNMCMD_TRACKS_IN_ROOT 0x48      // Anzahl Musikdateien im root zurückgeben, Param = 0


/***********************************************************************************************************************
* private Funktion
*
* Sendet ein Byte über den UART an das MP3-Modul
* Übergabe: Byte
***********************************************************************************************************************/
static void uart_send_byte (uint8_t byte)
{
  while ( !( UCSRA & (1<<UDRE)) );                          // ggf laufenden Versand abwarten
  UDR = byte;
}
/***********************************************************************************************************************
* private Funktion
*
* Empfängt ein Byte via UART vom MP3-Modul
***********************************************************************************************************************/
uint8_t uart_rec_byte (void)
{
  while ( !(UCSRA & (1<<RXC)) );                            // Warten auf Byte
  return UDR;
}
/***********************************************************************************************************************
* private Funktion
*
* sendet Befehl an das MP3-Modul via UART
* Dazu werden 10 Bytes versandt die wie unten zugeordnet sind
* inlusive ist eine Prüfsumme
***********************************************************************************************************************/
static void fnm_send_cmd (uint8_t cmd, uint8_t feedback, uint16_t param)
{
  uint16_t  chk;
  uint8_t   cnt;
  uint8_t   buffer[10];

  buffer[0] = 0x7E;                                         // $S Start byte 0x7E
  buffer[1] = 0xFF;                                         // Version 0xFF by default
  buffer[2] = 6;                                            // Length (Version...ParamLSB) = 6 Bytes
  buffer[3] = cmd;                                          // CMD
  buffer[4] = feedback;                                     // 1: Need feedback, 0: No need feedback
  buffer[5] = (uint8_t)((param >> 8) & 0xFF);               // Param_MSB
  buffer[6] = (uint8_t)(param & 0xFF);                      // Param_LSB
  chk = (uint16_t)buffer[1] + (uint16_t)buffer[2] +         // Checksumme
        (uint16_t)buffer[3] + (uint16_t)buffer[4] +         // berechnen
        (uint16_t)buffer[5] + (uint16_t)buffer[6];
  chk = 0 - chk;
  buffer[7] = (uint8_t)((chk >> 8) & 0xFF);                 // Checksumme
  buffer[8] = (uint8_t)(chk & 0xFF);                        // mit senden
  buffer[9] = 0xEF;                                         // End byte 0xEF

  for (cnt=0; cnt<=9; cnt++) uart_send_byte(buffer[cnt]);   // 10 Byte Telegramm versenden
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* Initialisiert die IO-Pins (MP3-Modul angeschlossen über UART)
* Datenübertragung 8N1 mit 9600 Baud
***********************************************************************************************************************/
void mp3_player_init (void)
{
  PORTC |= 0b00000010;                                      // PD1 = TX = Ausgang
  UBRRH = 0;                                                // bei 7,3728 MHz
  UBRRL = 47;                                               // Baudrate = 9600
  UCSRB = (1<<RXEN)|(1<<TXEN);                              // Receiver undTransmitter einschalten
  UCSRC = (1<<URSEL)|(1 << UCSZ1)|(1 << UCSZ0);             // Asynchron 8N1
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* Stopt laufende Wiedergabe
***********************************************************************************************************************/
void mp3_player_stop (void)
{
  mp3_player_set_volume(0);
  fnm_send_cmd(FNMCMD_STOP, 0, 0);
  _delay_ms(50);
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* Stopt eine laufende Wiedergabe und versetzt Gerät in Sleep Mode
***********************************************************************************************************************/
void mp3_player_sleep (void)
{
  mp3_player_stop();
  mp3_player_set_volume(0);
  fnm_send_cmd(FNMCMD_SLEEP, 0, 0);
  _delay_ms(50);
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* Sendet Soft-Reset Befehl an MP3-Player und setzt
* Device auf SD-Card, den EQ auf Stufe Normal
***********************************************************************************************************************/
void mp3_player_wakeup (void)
{
  fnm_send_cmd(FNMCMD_RESET, 0, 0);
  _delay_ms(2000);
  fnm_send_cmd(FNMCMD_SET_DEVICE, 0, 2);
  _delay_ms(300);
  fnm_send_cmd(FNMCMD_SET_EQ, 0, 0);
  _delay_ms(50);
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* startet mit Wiedergabe eines Stückes
* Übergabe: Titelnummer 1...MP3_TRACKS
***********************************************************************************************************************/
void mp3_player_play (uint8_t song_no)
{
  fnm_send_cmd(FNMCMD_SET_TRACK, 0, (uint16_t)song_no);
  _delay_ms(50);
  fnm_send_cmd(FNMCMD_PLAY, 0, 0);
  _delay_ms(50);
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* zum setzen der Ausgabelaustärke
***********************************************************************************************************************/
void mp3_player_set_volume (uint8_t volume)
{
  fnm_send_cmd(FNMCMD_SET_VOLUME, 0, (uint16_t)volume);
  _delay_ms(50);
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* liest Anzahl der abspielbaren Musikstücke im Wurzelverzeichnis der SD-Karte und gibt Anzahl zurück
* es dürfen max. 99 Musikdateien im root der Karte gespeichert werden
***********************************************************************************************************************/
uint8_t mp3_player_get_tracks (void)
{
  uint8_t buffer[10];
  uint8_t cnt;

  while ( UCSRA & (1<<RXC) ) cnt = UDR;                     // Empfangspuffer leeren
  fnm_send_cmd(FNMCMD_TRACKS_IN_ROOT, 1, 0);                // Anfrage senden (und feedback anfordern)
  for (cnt=0; cnt<10; cnt++) buffer[cnt] = uart_rec_byte(); // Antwort -> 10 Bytes Frame vom Modul abholen
  return buffer[6];                                         // Anzahl in Byte6
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* sendet Befehl zum Lautstärke erhöhen an MP3-Player
***********************************************************************************************************************/
void mp3_player_inc_volume (void)
{
  fnm_send_cmd(FNMCMD_INC_VOLUME, 0, 0);
}
/***********************************************************************************************************************
* öffentliche Funktion
*
* sendet Befehl zum Lautstärke verringern an MP3-Player
***********************************************************************************************************************/
void mp3_player_dec_volume (void)
{
  fnm_send_cmd(FNMCMD_DEC_VOLUME, 0, 0);
}
