/*
 * ds3231.c
 *
 * Copyright 2018 Thomas Gollmer <th_goso@freenet.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 * I2C-Low-Level-Funktionen DS3231-Board
 */

#include <util/twi.h>
#include "rtc.h"

//I2C Adresse DS3231
#define DS_I2CADR_R   0b11010001    // Für Lesezugriffe
#define DS_I2CADR_W   0b11010000    // Für Schreibzugriffe
// Benötigte DS Register/Adressen für Funktionsübergaben
#define DS_SECONDS    0x00
#define DS_MINUTES    0x01
#define DS_HOURS      0x02
#define DS_CONTROL    0x0E

/***********************************************************************************************************************
 * Private Funktion
 * zum Wandeln von PackedBCD "00" bis "99" nach binär
 **********************************************************************************************************************/
static uint8_t pbcd_to_bin (uint8_t pbcd)
{
  uint8_t high;

  high = ((pbcd >> 4) & 0b00001111) * 10;
  pbcd &= 0b00001111;

  return high + pbcd;
}
/***********************************************************************************************************************
 * Private Funktion
 * zum Wandeln von 0 bis 99 nach PackedBCD "00" bis "99"
 **********************************************************************************************************************/
static uint8_t bin_to_pbcd (uint8_t bin)
{
  uint8_t high = 0;

  while (bin >= 10) {
    high++;
    bin -= 10;
  }

  return (high << 4) | bin;
}
/***********************************************************************************************************************
 * Private Funktion
 *
 * Liest ein Register des DS3231 aus und gibt es zurück
 * Übergeben werden muß Registernummer bzw. der Name aus "ds3231.h"
 * Die Daten vom DS liegen IMMER im PACKED BCD FORMAT vor
 * und werden vor der Rückgabe nach bin gewandelt
 **********************************************************************************************************************/
static uint8_t ds_read_reg (uint8_t reg)
{
  uint8_t ret;

  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);                   // Startbedingung
  while (!(TWCR & (1<<TWINT)));                             // senden
  TWDR = DS_I2CADR_W;                                       // Schreibzugriff
  TWCR = (1<<TWINT) | (1<<TWEN);                            // DS3231
  while (!(TWCR & (1<<TWINT)));
  TWDR = reg;                                               // Registeradresse senden
  TWCR = (1<<TWINT) | (1<<TWEN);
  while (!(TWCR & (1<<TWINT)));
  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);                   // Startbedingung
  while (!(TWCR & (1<<TWINT)));                             // wiederholen
  TWDR = DS_I2CADR_R;                                       // Lesezugriff
  TWCR = (1<<TWINT) | (1<<TWEN);                            // DS3231
  while (!(TWCR & (1<<TWINT)));
  TWCR = (1<<TWINT)|(1<<TWEN);                              // NACK senden / Byte empfangen
  while (!(TWCR & (1<<TWINT)));
  ret = TWDR;
  while(!(TWCR & (1<<TWINT)));
  TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);                   // Stopbedingung
  while (TWCR & (1<<TWINT));

  ret = pbcd_to_bin(ret);                                   // Rückgabe als Binär
  return ret;
}
/***********************************************************************************************************************
 * Private Funktion
 *
 * Schreibt "Wert" in das Register "Register" des DS
 * übergebener Wert muß wie für den DS üblich im PACKED BCD FORMAT vorliegen
 * "Register" muß Registernummer oder Bezeichner aus "ds3231.h" sein
 **********************************************************************************************************************/
static void ds_write_reg (uint8_t reg, uint8_t value)
{
  value = bin_to_pbcd(value);                               // Vorm Senden nach PBCD wandeln

  TWCR = (1<<TWINT)|(1<<TWSTA)|(1<<TWEN);                   // Startbedingung
  while (!(TWCR & (1<<TWINT)));                             // senden
  TWDR = DS_I2CADR_W;                                       // Schreibzugriff
  TWCR = (1<<TWINT) | (1<<TWEN);                            // DS3231
  while (!(TWCR & (1<<TWINT)));
  TWDR = reg;                                               // Registeradresse senden
  TWCR = (1<<TWINT) | (1<<TWEN);
  while (!(TWCR & (1<<TWINT)));
  TWDR = value;                                             // Wert senden
  TWCR = (1<<TWINT) | (1<<TWEN);
  while (!(TWCR & (1<<TWINT)));
  TWCR = (1<<TWINT)|(1<<TWEN)|(1<<TWSTO);                   // Stopbedingung
  while (TWCR & (1<<TWINT));
}
/***********************************************************************************************************************
 * Öffentliche Funktion
 *
 * Initialisiert die Ein/Ausgänge für den I2C-Bus (DS3231)
 * PC5/PC4 werden für I2C genutzt
 **********************************************************************************************************************/
void rtc_init (void)
{
  TWBR = 0b00000010;                                        // I2C Bitratenregister I2C Frequenz etwas unter 400kHz
  ds_write_reg(DS_CONTROL, 0);                              // Oszillator bei Batteriebetrieb weiter laufen lassen
}
/***********************************************************************************************************************
 * Öffentliche Funktion
 *
 * Speichern der übergebenen Uhrzeit in den Zeitregistern des DS3231
 **********************************************************************************************************************/
void rtc_set_time (timehms_t timehms)
{
  ds_write_reg(DS_HOURS, timehms.h);
  ds_write_reg(DS_MINUTES, timehms.m);
  ds_write_reg(DS_SECONDS, timehms.s);
}
/***********************************************************************************************************************
 * Öffentliche Funktion
 *
 * Lesen der Zeitregister aus dem DS3231 und zurückgeben
 **********************************************************************************************************************/
timehms_t rtc_get_time (void)
{
  timehms_t retval;

  retval.h = ds_read_reg(DS_HOURS);
  retval.m = ds_read_reg(DS_MINUTES);
  retval.s = ds_read_reg(DS_SECONDS);

  return retval;
}
